#ifndef FEMDOF_H
#define FEMDOF_H

#include "fem.h"
#include <map>

namespace fem
{

/// a dof type (an index in the dofs table) e.g. "X"=>0, "Y"=>1, etc.

class FEM_API DOF
{
    DOFs const &dofs; ///< set of flags
    size_t idx;       ///< readonly index
public:
    size_t index() const { return idx; }
    friend FEM_API std::ostream &operator<<(std::ostream &out, DOF const &obj);

private:
    friend class DOFs;    // DOF must be created by DOFs
    friend class DOFMask; // used for accessing dofs (for printing)
    DOF(DOFs const &_dofs, size_t _idx) : dofs(_dofs), idx(_idx) {}
};

/// stores all dof types
///      this is the only class which is able to create a new type of dof

class FEM_API DOFs
{
    std::vector<std::string> names;
    std::map<std::string, size_t> idx_by_name;

public:
    DOFs();
    ~DOFs() {}
    size_t size() const { return names.size(); }
    std::string name(DOF const &dof) const { return names.at(dof.idx); }

    friend FEM_API std::ostream &operator<<(std::ostream &out, DOFs const &obj);
    void add(std::string const &dofname);
    DOF operator[](std::string const &dofname) const { return DOF(*this, idx_by_name.at(dofname)); }
    DOF operator[](char const *dofname) const { return DOF(*this, idx_by_name.at(std::string(dofname))); }
    DOF operator[](size_t idx) const { return DOF(*this, idx); }
};

/// a mask of dofs
///     allows the user to select several dof types using |

class FEM_API DOFMask
{
    DOFs const &dofs; ///< set
    size_t v;         ///< mask
public:
    /// create a mask from a single dof (implicitly if needed)
    DOFMask(DOF const &dof) : dofs(dof.dofs), v(1ULL << dof.index()) {}
    /// combine masks & other dofs
    DOFMask operator|(DOFMask const &obj) const
    {
        if (&dofs != &obj.dofs)
            throw std::runtime_error("incompatible sets"); // use assert?
        return DOFMask(dofs, v | obj.v);
    }
    /// check a mask versus another one
    bool operator&(DOFMask const &obj) const
    {
        if (&dofs != &obj.dofs)
            throw std::runtime_error("incompatible sets"); // use assert?
        return v & obj.v;
    }
    /// pretty print
    friend FEM_API std::ostream &operator<<(std::ostream &out, DOFMask const &obj);

private:
    /// rebuild a mask from the result of a boolean operation
    DOFMask(DOFs const &_dofs, size_t _v) : dofs(_dofs), v(_v) {}
};

inline DOFMask operator|(DOF const &d1, DOF const &d2)
{
    return (DOFMask)d1 | d2;
}

} // namespace fem

#endif // FEMDOF_H
